import logging
import os

from .auth.auth import Auth
from .auth.rsa_auth import SymBotRSAAuth
from .configure.configure import SymConfig


def load_from_env_var(env_var, delimiter=":"):
    """Look for an environment variable with the format:
        env_var=[RSA|CERT]:/path/to/config
    
    E.g. SYMPHONY_CONFIG=RSA:resources/config.json

    The env_var parameter describes which environment variable to check

    config.configure() and auth.authenticate() are called within this method so don't need to
    be called again

    Returns a tuple of configuration, auth
    """
    env_value = os.environ.get(env_var)
    if env_value is None:
        raise ValueError("Unable to find environment variable at: " + env_var)

    split = env_value.split(delimiter)

    if len(split) == 1:
        raise ValueError(f"Did not find {delimiter} in environment variable: {env_value}")
    elif len(split) > 2:
        # On windows you can have a colon in the path
        if len(split) == 3 and os.name == "nt":
            split = [split[0], split[1] + ":" + split[2]]
        else:
            raise ValueError(f"Found more than one {delimiter} in environment_variable: {env_value}") 
    
    
    if split[0].lower() not in ["rsa", "cert"]:
        raise ValueError(f"Didn't recognise f{split[0]}, expected one of: RSA, CERT")
    
    logging.debug(f"Loading config from {split[1]} with authentication mode: {split[0]}")

    conf = SymConfig(split[1], split[1])
    conf.load_config()

    if split[0].lower() == "rsa":
        auth = SymBotRSAAuth(conf)
    else:
        auth = Auth(conf)
    
    auth.authenticate()

    return conf, auth


def configure_logging(filename=None, directory=None, log_level=logging.DEBUG, format=None, filemode='a'):
    """Set up the loggers with basic defaults

        Set filename and directory to both be None to not save to file.

        filename: The filename to give to the logs. If this has paths as well they will be relative
                  to the directory. Defaults to log.log.
        directory: Directory to save the logs in. Defaults to a directory called "logs" adjacent
                   to the main. Put "." to make this current working directory. This will creat the
                   folder if it doesn't already exist.
        log_level: One of logging.levels
        format: Defaults to '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        filemode: How to open the logfile, 'w' rewrites, 'a' appends to existing
    """
    file_log = True
    if directory is None and filename is None:
        file_log = False

    if directory is None:
        # It's generally bad practice to import in functions, but in this instance
        # this import might fail (if it's loaded from a REPL etc) so it's probably
        # best to leave it until it's definitely needed.
        import __main__
        directory = os.path.dirname(__main__.__file__)
    elif directory == ".":
        directory = sys.curdir
    elif filename is None:
        filename = "log.log"
    
    if not file_log:
        full_path = None        
    else:
        full_path = os.path.join(directory, filename)
        # This has to be done again to handle the case where filename was relative
        full_dir = os.path.dirname(full_path)

        if not os.path.exists(full_dir):
            os.makedirs(full_dir, exist_ok=True)
            
    if format is None:
        format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

    logging.basicConfig(
            filename=full_path,
            format=format,
            filemode=filemode, level=log_level
    )
    logging.getLogger("urllib3").setLevel(logging.WARNING)



from .elements_listener import ElementsActionListener
import logging


# A sample implementation of Abstract imListener class
# The listener can respond to incoming events if the respective event
# handler has been implemented


class ElementsListenerTestImp(ElementsActionListener):
    """Example implementation of ElementsActionListener

        sym_bot_client: contains clients which respond to incoming events

    """

    def __init__(self, sym_bot_client):
        self.bot_client = sym_bot_client

    def on_elements_action(self, action):
        logging.debug('element submitted :')


class AsyncElementsListenerTestImp(ElementsActionListener):
    """Example implementation of ElementsListener with asynchronous functionality"""

    def __init__(self, sym_bot_client):
        self.bot_client = sym_bot_client

    async def on_elements_action(self, action):
        logging.debug('async element submitted :')



from yattag import Doc


class MessageFormatter:

    def __init__(self):
        pass

    def format_message(self, message):
        """
        appends messageML tags to plain text and returns a dictionary:

        {message : messageML object}
        """
        doc, tag, text, line = Doc().ttl()
        with tag('messageML'):
            text(message)

        return dict(message=doc.getvalue())



from ..processors.sym_message_parser import SymMessageParser
import logging


class MessageProcessor:
    def __init__(self, bot_client):
        self.bot_client = bot_client
        self.message_parser = SymMessageParser()

    def process(self, msg):
        logging.debug('inside of process')
        msg_text = self.message_parser.get_text(msg)
        msg_to_send = dict(
            message='<messageML>Hello {}, hope you are doing well!</messageML>'.format(
                self.message_parser.get_im_first_name(msg))
        )

        if msg_text:
            stream_id = self.message_parser.get_stream_id(msg)
            self.bot_client.get_message_client(). \
                send_msg(stream_id, msg_to_send)



from abc import ABC, abstractmethod

# Abstract class for connections listener
# class is just an interface of functions to handle the connection events
# received from DataFeed Connection Events are "CONNECTIONREQUESTED" and
# "CONNECTIONACCEPTED" handle actual event logic in your implementation of
# this abstract class


class ConnectionListener(ABC):

    @abstractmethod
    def on_connection_accepted(self, user):
        pass

    @abstractmethod
    def on_connection_requested(self, user):
        pass



import setuptools


def readme():
    with open('README.md') as f:
        return f.read()


setuptools.setup(
    name="sym_api_client_python",

    version="1.3.3",
    author="Symphony Platform Solutions",
    author_email="platformsolutions@symphony.com",
    description="Symphony REST API - Python Client",
    long_description=readme(),
    long_description_content_type='text/markdown',
    url="https://github.com/SymphonyPlatformSolutions/"
        "symphony-api-client-python",
    packages=setuptools.find_packages(),
    install_requires=[
        'aiohttp',
        'aioresponses>=0.6.1',
        'pyOpenSSL',
        'rsa',
        'requests',
        'python-jose~=3.2.0',
        'python-json-logger==0.1.11',
        'beautifulsoup4==4.8.0',
        'Jinja2==2.10.1',
        'requests_pkcs12==1.9',
        'requests-toolbelt==0.9.1',
        'requests-mock>=1.7.0',
        'yattag==1.12.2',
        'defusedxml==0.6.0'
    ],
    tests_require=['pytest'],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from .api_client import APIClient
import logging
import json

class DataFeedClientV2(APIClient):
    def __init__(self, bot_client):
        self.bot_client = bot_client
        self.ackid = ""

    def create_datafeed(self):
        url = '/agent/v5/datafeeds'
        response = self.bot_client.execute_rest_call("POST", url)

        datafeed_id = response.get("id")
        logging.debug('DataFeedClientV2/create_datafeed() --> {}'.format(datafeed_id))
        return datafeed_id


    def read_datafeed(self, datafeed_id, *ackId):
        """
        DF 2 Version: We need to use an ack Id to make sure the events are well received
        """
        logging.debug('DataFeedClientV2/read_datafeed()')
        url = '/agent/v5/datafeeds/{0}/read'.format(datafeed_id)
        data = {}
        if len(ackId) == 0:
            data["ackId"] = ""
        else:
            data["ackId"] = ackId[0]

        datafeed_read = self.bot_client.execute_rest_call("POST", url,  json=data)
        self.ackid = datafeed_read.get("ackId")
        events = datafeed_read.get("events")
        return events

    def list_datafeed_id(self):
        logging.debug('DataFeedClientV2/list_datafeed()')
        url = '/agent/v5/datafeeds'
        datafeed_ids = self.bot_client.execute_rest_call("GET", url)
        return datafeed_ids

    def delete_datafeed(self, datafeed_id):
        logging.debug('DataFeedClientV2/delete_datafeed()')
        url = '/agent/v5/datafeeds/{0}'.format(datafeed_id)
        self.bot_client.execute_rest_call("DELETE", url)

    def get_ack_id(self):
        return self.ackid


import logging

from .api_client import APIClient


class HealthCheckClient(APIClient):
    def __init__(self, bot_client):
        self.bot_client = bot_client

    def get_health_check(self, show_firehose_errors=False):
        logging.debug('HealthCheckClient/get_health_check()')
        params = {'showFirehoseErrors': show_firehose_errors}
        url = '/agent/v2/HealthCheck'
        return self.bot_client.execute_rest_call('GET', url, params=params)

    def ensure_all_services_up(self, check_firehose=False, fields_to_check=None):
        logging.debug('HealthCheckClient/ensure_all_services_up()')
        # This list would have to be updated if new fields became available in the health check
        if fields_to_check is None:
            fields_to_check = [
                'podConnectivity',
                'keyManagerConnectivity',
                'encryptDecryptSuccess',
                'agentServiceUser',
                'ceServiceUser',
            ]
        if check_firehose:
            fields_to_check.append('firehoseConnectivity')

        health_check = self.get_health_check(check_firehose)

        logging.debug(health_check)

        functioning = [ health_check[field] for field in fields_to_check ]
        if not all(functioning):
            problems = [fields_to_check[i] for i,v in enumerate(functioning) if not v]
            raise RuntimeError(f"Not all services available: {problems}")
            




import logging

from .api_client import APIClient


# child class of APIClient --> Extends error handling functionality
# ConnectionsClient class contains a series of functions corresponding to all
# pod admin endpoints on the REST API.
class ConnectionsClient(APIClient):

    def __init__(self, bot_client):
        self.bot_client = bot_client

    def create_connection(self, user_id):
        """
        Pods from all users involved need to have crossPod enabled between them.

        Users who belong to the same private pod are implicitly connected.
        If you attempt to connect with an internal user, this endpoint will
        return the corresponding connection object with a status of accepted
        """
        logging.debug('ConnectionsClient/create_connection()')
        url = '/pod/v1/connection/create'
        data = {'userId': user_id}
        return self.bot_client.execute_rest_call('POST', url, json=data)

    def get_connection(self, user_id):
        """
        When calling this as an OBO-enabled endpoint,
        use the OBO User Authenticate token for sessionToken
        """
        logging.debug('ConnectionsClient/get_connection()')
        url = '/pod/v1/connection/user/{0}/info'.format(user_id)
        return self.bot_client.execute_rest_call('GET', url)

    def list_connections(self, status, **kwargs):
        """
        This retrieves all connections of the requesting user.
        (i.e. both connections in which the requesting user is the sender and
        those in which the requesting user is the inivtee).
        By default, if you haven't specified the connection status to filter on,
         this call will only return results for both "pending_incoming" and
         "pending_outgoing".
        """
        logging.debug('ConnectionsClient/list_connections()')
        url = '/pod/v1/connection/list'
        params = {'status' : status}
        return self.bot_client.execute_rest_call('GET', url, params=params)

    def accept_connection(self, user_id):
        """
        This allows the user to accept a specific connection request.
        For users of the same private pod who are implicitly connected,
        this endpoint returns the connection with status of "Accepted".
        """
        logging.debug('ConnectionsClient/accept_connection()')
        url = '/pod/v1/connection/accept'
        data = {'userId': user_id}
        return self.bot_client.execute_rest_call('POST', url, json=data)

    def reject_connection(self, user_id):
        """
        This allows the user to reject a specific connection request.
        Reject the connection between the requesting user and request sender.
        If both users are in the same private pod, an error will be returned
        because both users have an implicit connection which cannot be rejected.
        """
        logging.debug('ConnectionsClient/reject_connection()')
        url = '/pod/v1/connection/reject'
        data = {'userId': user_id}
        return self.bot_client.execute_rest_call('POST', url, json=data)

    def remove_connection(self, user_id):
        """
        Removes a connection with a user.
        This endpoint returns 400 Bad Request when the users aren't connected.
        For example, when a user hasn’t yet accepted a connection request
        from another user.
        """
        logging.debug('ConnectionsClient/remove_connection()')
        url = '/pod/v1/connection/user/{0}/remove'.format(user_id)
        return self.bot_client.execute_rest_call('POST', url)




from .connection_listener import ConnectionListener
import logging

# sample implementation of Abstract imListener class
# has instance of SymBotClient so that it can respond to events coming in by
# leveraging other clients on SymBotClient
# each function should contain logic for each corresponding event


class ConnectionListenerTestImp(ConnectionListener):
    """Example implementation of ConnectionListener

        sym_bot_client: contains clients which respond to incoming events

    """

    def __init__(self, sym_bot_client):
        self.bot_client = sym_bot_client

    def on_connection_accepted(self, connection):
        logging.debug('on_connection_accepted()', connection)

    def on_connection_requested(self, connection):
        logging.debug('on_connection_requested', connection)



class DatafeedExpiredException(Exception):
    """This exception represents the response below from the API that occurs periodically
    when a datafeed has expired. It should be handled by creating a new datafeed and 
    reading from there"""
    pass



class SymElementsParser:
    """
    FormParser class takes elements_action_data as its only parameter.
    elements action data is a json object passed from DataFeedEventService
    to the ElementsListenerImplementation class.  The elements_action_data json
    follows the structure below:

    [
        {
            "id": "kzwn0A",
            "messageId": "Yzy7OoPQnLI9DTCKWmIRMX___pQEChSZbQ",
            "timestamp": 1563300326246,
            "type": "SYMPHONYELEMENTSACTION",
            "initiator": {
                "user": {
                    "userId": 7078106482890,
                    "firstName": "User",
                    "lastName": "Bot",
                    "displayName": "User",
                    "email": "user_bot@symphony.com",
                    "username": "user_bot"
                }
            },
            "payload": {
                "symphonyElementsAction": {
                    "stream": {
                        "streamId": "0YeiA-neZa1PrdHy1L82jX___pQjntU-dA",
                        "streamType": "IM"
                    },
                    "formId": "form_id",
                    "formValues": {
                        "action": "submit_button",
                        "name_01": "John Doe",
                        "country": "opt1",
                        "example_radio": "option_01",
                        "checkbox_1": "value01",
                        "checkbox_2": "value02",
                        "comment": "In my opinion..."
                    }
                }
            }
        }
    ]

    This Class provides methods to the developer to easily access data in this
    elements action payload corresponding to each element in the form:
    """

    def __init__(self):
        pass

    def get_stream_id(self, elements_action_data):
        if 'stream' in elements_action_data['payload']['symphonyElementsAction'] and 'streamId' in \
                elements_action_data['payload']['symphonyElementsAction']['stream']:
            return elements_action_data['payload']['symphonyElementsAction']['stream']['streamId']
        if 'formStream' in elements_action_data['payload']['symphonyElementsAction'] and 'streamId' in \
                elements_action_data['payload']['symphonyElementsAction']['formStream']:
            return elements_action_data['payload']['symphonyElementsAction']['formStream']['streamId'].rstrip(
                '=').replace('/', '_').replace('+', '-')

    def get_stream_type(self, elements_action_data):
        if 'stream' in elements_action_data['payload']['symphonyElementsAction'] and 'streamType' in \
                elements_action_data['payload']['symphonyElementsAction']['stream']:
            return elements_action_data['payload']['symphonyElementsAction']['stream']['streamType']

    def get_action_stream_id(self, elements_action_data):
        """
        returns streamId for the actionStream.
        actionStream is where the formMessage is sent, this is a hidden stream
        so that the encryption context is between user and bot only and not everyone else on the room
        """
        if elements_action_data['payload']['symphonyElementsAction']['actionStream'] != None and \
                elements_action_data['payload']['symphonyElementsAction']['actionStream']['streamId'] != None:
            return elements_action_data['payload']['symphonyElementsAction']['actionStream']['streamId']

    def get_form_message_id(self, elements_action_data):
        """
        returns formMessageId corresponding to the element sent:
        """
        return elements_action_data['payload']['symphonyElementsAction']['formMessageId']

    def get_form_id(self, elements_action_data):
        """
        returns the unique formId assosiated with this form element
        """
        return elements_action_data['payload']['symphonyElementsAction']['formId']

    def get_form_values(self, elements_action_data):
        """
        returns a dictionary of key value pairs in the following format:
        {
            "element_name": "element_value"
        }
        For Example:
        {
            "action": "submit_button",
            "name_01": "John Doe",
            "country": "opt1",
            "example_radio": "option_01",
            "checkbox_1": "value01",
            "checkbox_2": "value02",
            "comment": "In my opinion..."
        }
        """

        return elements_action_data['payload']['symphonyElementsAction']['formValues']

    def get_action(self, elements_action_data):
        """
        returns name of button that submitted the form when clicked
        """
        return elements_action_data['payload']['symphonyElementsAction']['formValues']['action']

    def get_initiator_user_id(self, elements_action_data):
        """
        returns id of user that submitted the form
        """
        return elements_action_data['initiator']['user']['userId']

    def get_initiator_display_name(self, elements_action_data):
        """
        returns displayName of user that submitted the form
        """
        return elements_action_data['initiator']['user']['displayName']



from abc import ABC, abstractmethod

# Abstract class for ROOM listener
# class is just an interface of functions to handle the Room events received
# from DataFeed see Real Time Events in REST API documentation for more
# details. The developer will handle actual event logic in your implementation
# of this abstract class


class RoomListener(ABC):

    @abstractmethod
    def on_room_msg(self, message):
        pass

    @abstractmethod
    def on_room_created(self, roomCreated):
        pass

    @abstractmethod
    def on_room_deactivated(self, roomDeactivated):
        pass

    @abstractmethod
    def on_room_member_demoted_from_owner(self, roomMemberDemotedFromOwner):
        pass

    @abstractmethod
    def on_room_member_promoted_to_owner(self, roomMemberPromotedToOwner):
        pass

    @abstractmethod
    def on_room_reactivated(self, roomReactivated):
        pass

    @abstractmethod
    def on_room_updated(self, roomUpdated):
        pass

    @abstractmethod
    def on_user_joined_room(self, userJoinedRoom):
        pass

    @abstractmethod
    def on_user_left_room(self, userLeftRoom):
        pass



import asyncio
import uuid
from functools import partial
import logging
import datetime
from collections import namedtuple

from asyncio import CancelledError

#TODO: These imports are duplicated over the abstract class to avoid errors in the Async version
from .exceptions.UnauthorizedException import UnauthorizedException
from .exceptions.APIClientErrorException import APIClientErrorException
from .exceptions.DatafeedExpiredException import DatafeedExpiredException
from .exceptions.ServerErrorException import ServerErrorException
from .exceptions.MaxRetryException import MaxRetryException

from .services.abstract_datafeed_event_service import AbstractDatafeedEventService
from .clients.constants.DatafeedVersion import DatafeedVersion
from .services.datafeed_event_service_v1 import DataFeedEventServiceV1
from .services.datafeed_event_service_v2 import DataFeedEventServiceV2

log = logging.getLogger(__name__)


def make_datetime(unix_timestamp_millis):
    seconds, millis = divmod(unix_timestamp_millis, 1000)
    return datetime.datetime.utcfromtimestamp(seconds).replace(microsecond=millis * 1000)


EventTrace = namedtuple('EventTrace', 'message_id creation_time bot_received listeners_complete bot_complete')


class DataFeedEventService:

    def __init__(self, sym_bot_client, error_timeout_sec=None, maximum_timeout_sec=None):
        config = sym_bot_client.get_sym_config()

        # Creating the DataFeed Event Service
        if DatafeedVersion.version_of(config.data.get("datafeedVersion")) == DatafeedVersion.V2:
            self.datafeed_event_service = DataFeedEventServiceV2(sym_bot_client, error_timeout_sec=error_timeout_sec,
                                                                 maximum_timeout_sec=maximum_timeout_sec)
        else:
            self.datafeed_event_service = DataFeedEventServiceV1(sym_bot_client, error_timeout_sec=error_timeout_sec,
                                                                 maximum_timeout_sec=maximum_timeout_sec)

    def start_datafeed(self):
        """Start reading events from datafeed.
        """
        self.datafeed_event_service.start_datafeed()

    def read_datafeed(self):
        self.datafeed_event_service.read_datafeed()

    def activate_datafeed(self):
        self.datafeed_event_service.activate_datafeed()

    def deactivate_datafeed(self):
        self.datafeed_event_service.deactivate_datafeed()

    def add_room_listener(self, room_listener):
        self.datafeed_event_service.add_room_listener(room_listener)

    def remove_room_listener(self, room_listener):
        self.datafeed_event_service.remove_room_listener(room_listener)

    def add_im_listener(self, im_listener):
        self.datafeed_event_service.add_im_listener(im_listener)

    def remove_im_listener(self, im_listener):
        self.datafeed_event_service.remove_im_listener(im_listener)

    def add_elements_listener(self, elements_listener):
        self.datafeed_event_service.add_elements_listener(elements_listener)

    def remove_elements_listener(self, elements_listener):
        self.datafeed_event_service.remove_elements_listener(elements_listener)

    def add_connection_listener(self, connection_listener):
        self.datafeed_event_service.add_connection_listener(connection_listener)

    def remove_connection_listener(self, connection_listener):
        self.datafeed_event_service.remove_connection_listener(connection_listener)

    def add_wall_post_listener(self, wall_post_listener):
        self.datafeed_event_service.add_wall_post_listener(wall_post_listener)

    def remove_wall_post_listener(self, wall_post_listener):
        self.datafeed_event_service.remove_wall_post_listener(wall_post_listener)

    def add_suppression_listener(self, suppression_listener):
        self.datafeed_event_service.add_suppression_listener(suppression_listener)

    def remove_suppression_listener(self, suppression_listener):
        self.datafeed_event_service.remove_suppression_listener(suppression_listener)

    def handle_events(self, events):
        self.datafeed_event_service.handle_events(events)

    def handle_event(self, payload):
        self.datafeed_event_service.handle_event(payload)

    ### Handlers ###
    def msg_sent_handler(self, payload):
        self.datafeed_event_service.msg_sent_handler(payload)

    def instant_msg_handler(self, payload):
        self.datafeed_event_service.instant_msg_handler(payload)

    def room_created_handler(self, payload):
        self.datafeed_event_service.room_created_handler(payload)

    def room_updated_handler(self, payload):
        self.datafeed_event_service.room_updated_handler(payload)

    def room_deactivated_handler(self, payload):
        self.datafeed_event_service.room_deactivated_handler(payload)

    def room_reactivated_handler(self, payload):
        self.datafeed_event_service.room_reactivated_handler(payload)

    def user_joined_room_handler(self, payload):
        self.datafeed_event_service.user_joined_room_handler(payload)

    def user_left_room_handler(self, payload):
        self.datafeed_event_service.user_left_room_handler(payload)

    def promoted_to_owner(self, payload):
        self.datafeed_event_service.promoted_to_owner(payload)

    def demoted_from_owner(self, payload):
        self.datafeed_event_service.demoted_from_owner(payload)

    def connection_accepted_handler(self, payload):
        self.datafeed_event_service.connection_accepted_handler(payload)

    def connection_requested_handler(self, payload):
        self.datafeed_event_service.connection_requested_handler(payload)

    def elements_action_handler(self, payload):
        self.datafeed_event_service.elements_action_handler(payload)

    def shared_post_handler(self, payload):
        self.datafeed_event_service.shared_post_handler(payload)

    def suppressed_message_handler(self, payload):
        self.datafeed_event_service.suppressed_message_handler(payload)

    ### Handle errors ###
    def get_and_increase_timeout(self, previous_exc=None):
        self.datafeed_event_service.get_and_increase_timeout(previous_exc)

    def decrease_timeout(self):
        self.datafeed_event_service.decrease_timeout()


class AsyncDataFeedEventService(AbstractDatafeedEventService):
    """Non-blocking datafeed event service.

    The main advantage this carries over DataFeedEventService is that read_datafeed is non-blocking
    and therefore a bot can listen to this feed and conduct other activities at the same time. This
    is achieved with async / await. It would typically be used as following:

        datafeed_event_service = bot_client.get_async_datafeed_event_service()
        datafeed_event_service.add_im_listener(async_listener)

        loop = asyncio.get_event_loop()
        loop.run_until_complete(datafeed_event_service.start_datafeed())

    There are several things to note about using this datafeed instead of the synchronous version:
        * All methods in the listener must be declared async def. Even if there is no async code
          inside them failing to declare them as such will likely cause an error when the service
          awaits them.
        * Listener methods that are expensive and not awaited will block reading the datafeed. For
          common IO-bound operations an asychronous version may be available, aiohttp instead of
          requests for example. If one is not, consider running it in a ThreadPoolExecutor
        * Some assumptions about ordering will fail. For example a user sending two messages to a
          bot in quick succession may get their responses in a different order.

    Potential improvements:
        * Provide a timeout to allow handlers to be cancelled after a certain period
        * Allow exception handling around listeners to be customised

    Known Issues:
        On Windows with Python 3.8+ there is a known issue with aiohttp and setting a proxy
        that can be found at https://github.com/aio-libs/aiohttp/issues/4536

        To work around this we must specify the event loop to use WindowsSelectorEventLoopPolicy with:
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        If cross-platform code is needed you can use a check:
        if sys.version_info[0] == 3 and sys.version_info[1] >= 8 and sys.platform.startswith('win'):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.queue = asyncio.Queue()
        self.exception_queue = asyncio.Queue()
        self.exception_handler = kwargs.pop('exception_handler', None)
        self.trace_enabled = kwargs.pop('trace_enabled', True)
        self.trace_recorder = kwargs.pop('trace_recorder', None)
        self.trace_dict = {}
        self.handle_events_task = None
        self.tasks = []
        self.datafeed_id = None

    async def start_datafeed(self):
        log.debug('AsyncDataFeedEventService/start_datafeed()')
        self.datafeed_id = self._get_from_file_or_create_datafeed_id()
        await asyncio.gather(self.read_datafeed(), self.handle_events(), self.handle_exceptions())

    async def deactivate_datafeed(self, wait_for_handler_completions=True):
        """Deactivating the datafeed may take up to 30 seconds while waiting for
        a 204 from the read_datafeed API"""
        log.debug('AsyncDataFeedEventService/deactivate_datafeed()')
        if wait_for_handler_completions:
            log.debug('AsyncDataFeedEventService/deactivate_datafeed() --> '
                     'Waiting for {} events to finish'.format(self.queue.qsize()))
            await self.queue.join()
            log.debug('AsyncDataFeedEventService/deactivate_datafeed() --> '
                     'Deactivating')
        else:
            log.debug('AsyncDataFeedEventService/deactivate_datafeed() --> '
                     '{} events still being handled, deactivating anyway'.format(self.queue.qsize()))

        if not self.stop:
            self.stop = True

        await self.queue.put(None)
        await self.exception_queue.put(None)
        await self.bot_client.close_async_sessions()

    async def read_datafeed(self):
        while not self.stop:
            try:
                events = await self.datafeed_client.read_datafeed_async(self.datafeed_id)
            except CancelledError as exc:
                log.info("Cancel request received. Stopping datafeed...")
                await self.deactivate_datafeed()
                raise
            except Exception as exc:
                try:
                    await self.handle_datafeed_errors(exc)
                except Exception as inner_exc:
                    await self.deactivate_datafeed(wait_for_handler_completions=False)
                    raise inner_exc from None
                continue

            self.decrease_timeout()
            if events:
                bot_id = self.bot_client.get_bot_user_info()['id']
                for event in events:
                    log.debug(
                        'AsyncDataFeedEventService/read_datafeed() --> '
                        'Incoming event from read_datafeed() with id: {}'.format(event.get('id'))
                    )

                    if event['initiator']['user']['userId'] != bot_id:
                        e_id = self._get_event_id(event)
                        self._add_trace(e_id, event["timestamp"])
                        await self.queue.put(event)
                    log.debug(f"Event queued. Current queue size: {self.queue.qsize()}")

            else:
                log.debug(
                    'AsyncDataFeedEventService() - no data coming in from '
                    'datafeed: {}'.format(self.datafeed_id)
                )

    async def handle_datafeed_errors(self, thrown_exception):
        """Various errors may get thrown by the datafeed reader, from 500s when a server node is
        being bounced, intermittent connectivity or SSL problems. With the exception of a 403
        UnauthorisedError where reauthentication occurs upstream, these are all handled in
        the same way - sleep, request a new datafeed id and then retry.
        """
        try:
            raise thrown_exception
        except UnauthorizedException:
            log.error('AsyncDataFeedEventService - caught unauthorized exception')
        except MaxRetryException as exc:
            log.error('AsyncDataFeedEventService - Bot has failed to authenticate more than 5 times ')
            raise exc from None

        except (DatafeedExpiredException, APIClientErrorException, ServerErrorException) as exc:
            log.error('AsyncDataFeedEventService - ' + str(exc))

        except Exception as exc:
            log.exception('AsyncDataFeedEventService - Unknown exception: ' + str(exc))

        sleep_for = self.get_and_increase_timeout(thrown_exception)
        log.debug('AsyncDataFeedEventService/handle_event() --> Sleeping for {:.4g}s'.format(sleep_for))
        await asyncio.sleep(sleep_for)
        try:
            log.debug('AsyncDataFeedEventService/handle_event() --> Restarting Datafeed')
            self.datafeed_id = self._create_datafeed_and_persist()
        except Exception as exc:
            await self.handle_datafeed_errors(exc)

    def _check_result(self, e_id, task):
        """Callback for task completion. If exceptions occurred add them for processing on the
        exceptions queue, otherwise they get swallowed by the future
        """
        self._add_trace(e_id)
        if task.exception() is not None:
            log.debug("Adding exception to exception queue for event: {}".format(e_id))
            self.exception_queue.put_nowait((e_id, task))
        else:
            self._process_full_trace(e_id)
        self.queue.task_done()

    def _process_full_trace(self, id):
        """Process all tracedata for the current request and append it to trace_recorder, if defined.

        Use with messagedId if available
        """

        if self.trace_enabled:
            try:
                intermediates = self.trace_dict[id]
                assert len(intermediates) == 4
                trace = EventTrace(id, intermediates[0], intermediates[1], intermediates[2],
                                   intermediates[3])
                total_time = intermediates[3] - intermediates[0]
                time_in_bot = intermediates[3] - intermediates[1]

                # This just writes out total seconds instead of formatting into minutes and hours
                # for a typical bot response this seems reasonable
                log.debug("Responded to message in: {:.4g}s. Including {:.4g}s inside the bot"
                         .format(total_time.total_seconds(), time_in_bot.total_seconds()))
                if self.trace_recorder is not None:
                    self.trace_recorder.append(trace)
                del self.trace_dict[id]
            except Exception as exc:
                log.error("Error while computing trace results for id: " + id)
                log.exception(exc)

    @staticmethod
    def _get_event_id(event):
        event_id = event["messageId"] if "messageId" in event else event.get('id')
        if event_id is None:
            event['id'] = str(uuid.uuid4())
            return event['id']
        return event_id

    def _add_trace(self, e_id, first_timestamp=None):
        """Take the current timestamp and add if to the trace_dict, used to trace the time taken
        to process events.

        Use with messageId if available
        """
        if self.trace_enabled:
            if first_timestamp is not None:
                self.trace_dict[e_id] = [make_datetime(first_timestamp)]
            try:
                self.trace_dict[e_id].append(datetime.datetime.utcnow())
            except KeyError:
                log.error(
                    'Error making traces for {}. Has the same messageId appeared'
                    'more than once?'.format(e_id)
                )

    async def handle_events(self):
        """For each event resolve its handler and add it to the queue to be processed"""
        while not self.stop:
            event = await self.queue.get()

            if event is None:
                # None is a sentinel to indicate the that work is done and there will be no more
                # messages. If there are multiple consumers of the queue, this will become an
                # issue as None will need to be added for each consumer
                self.queue.task_done()
            else:
                event_type = str(event['type'])
                e_id = self._get_event_id(event)
                self._add_trace(e_id)

                log.debug('AsyncDataFeedEventService/handle_events() --> event-type:' + event_type)
                try:
                    route = self.routing_dict[event_type]
                except KeyError:
                    log.debug('no event detected')
                    self.queue.task_done()
                    return

                future = asyncio.ensure_future(route(event))
                future.add_done_callback(partial(self._check_result, e_id))

    async def handle_exceptions(self):
        """If exceptions are not excplicitly handled they'll silently fail in the co-routine.
        This method picks results one by one off the queue and checks if they were successful, using
        the provided exception handler or failing otherwise"""
        while not self.stop:
            log.debug("Waiting on new exception, queue size: {}".format(self.exception_queue.qsize()))
            exception_event = await self.exception_queue.get()
            if exception_event is None:
                # Same as handle_events, None is a sentinel to close the queue and finish
                self.exception_queue.task_done()
            else:
                e_id, fut = exception_event
                try:
                    await fut
                except Exception as exc:
                    if self.exception_handler is not None:
                        self.exception_handler(exc)
                    else:
                        raise exc

                self._add_trace(e_id)
                self._process_full_trace(e_id)
                self.exception_queue.task_done()

    async def msg_sent_handler(self, payload):
        """This handler is used for both room messages and IMs. Which listener is invoked
        depends on the streamType"""
        log.debug('async msg_sent_handler function started')
        stream_type = payload['payload']['messageSent']['message']['stream']['streamType']
        message_sent_data = payload['payload']['messageSent']['message']
        if str(stream_type) == 'ROOM':
            for listener in self.room_listeners:
                await listener.on_room_msg(message_sent_data)
        else:
            for listener in self.im_listeners:
                await listener.on_im_message(message_sent_data)

    async def instant_msg_handler(self, payload):
        log.debug('async instant_msg_handler function started')
        instant_message_data = payload['payload']['instantMessageCreated']
        for listener in self.im_listeners:
            await listener.on_im_created(instant_message_data)

    async def room_created_handler(self, payload):
        log.debug('async room_created_handler function started')
        room_created_data = payload['payload']['roomCreated']
        for listener in self.room_listeners:
            await listener.on_room_created(room_created_data)

    async def room_updated_handler(self, payload):
        log.debug('async room_updated_handler')
        room_updated_data = payload['payload']['roomUpdated']
        for listener in self.room_listeners:
            await listener.on_room_updated(room_updated_data)

    async def room_deactivated_handler(self, payload):
        log.debug('async room_deactivated_handler')
        room_deactivated_data = payload['payload']['roomDeactivated']
        for listener in self.room_listeners:
            await listener.on_room_deactivated(room_deactivated_data)

    async def room_reactivated_handler(self, payload):
        log.debug('async room_reactivated_handler')
        room_reactivated_data = payload['payload']['roomReactivated']
        for listener in self.room_listeners:
            await listener.on_room_reactivated(room_reactivated_data)

    async def user_joined_room_handler(self, payload):
        log.debug('async user_joined_room_handler')
        user_joined_room_data = payload['payload']['userJoinedRoom']
        for listener in self.room_listeners:
            await listener.on_user_joined_room(user_joined_room_data)

    async def user_left_room_handler(self, payload):
        log.debug('async user_left_room_handler')
        user_left_room_data = payload['payload']['userLeftRoom']
        for listener in self.room_listeners:
            await listener.on_user_left_room(user_left_room_data)

    async def promoted_to_owner(self, payload):
        log.debug('async promoted_to_owner')
        promoted_to_owner_data = payload['payload']['roomMemberPromotedToOwner']
        for listener in self.room_listeners:
            await listener.on_room_member_promoted_to_owner(promoted_to_owner_data)

    async def demoted_from_owner(self, payload):
        log.debug('async demoted_from_owner')
        demoted_to_owner_data = payload['payload']['roomMemberDemotedFromOwner']
        for listener in self.room_listeners:
            await listener.on_room_member_demoted_from_owner(demoted_to_owner_data)

    async def connection_accepted_handler(self, payload):
        log.debug('async connection_accepted_handler')
        connection_accepted_data = payload['payload']['connectionAccepted']
        for listener in self.connection_listeners:
            await listener.on_connection_accepted(connection_accepted_data)

    async def connection_requested_handler(self, payload):
        log.debug('async connection_requested_handler')
        connection_requested_data = payload['payload']['connectionRequested']
        for listener in self.connection_listeners:
            await listener.on_connection_requested(connection_requested_data)

    async def elements_action_handler(self, payload):
        log.debug('async elements_action_handler')
        for listener in self.elements_listeners:
            await listener.on_elements_action(payload)

    async def shared_post_handler(self, payload):
        log.debug('shared_post_handler')
        shared_post = payload['payload']['sharedPost']
        for listener in self.wall_post_listeners:
            await listener.on_shared_post(shared_post)

    async def suppressed_message_handler(self, payload):
        log.debug('suppressed_message_handler')
        message_suppressed = payload['payload']['messageSuppressed']
        for listener in self.suppression_listeners:
            await listener.on_message_suppression(message_suppressed)
